package org.twodot.vnc;

import danger.ui.*;
import danger.app.Application;
import danger.app.Event;
import danger.util.DEBUG;

import java.io.IOException;

/**
 * 
 */
public class VirtualKeyboardDialog extends DialogWindow implements Resources, Commands{

    private rfbProto _rfb;
    private CheckBox _chkShift;
    private CheckBox _chkControl;
    private CheckBox _chkAlt;
    private CheckBox _chkMeta;

    private RadioButton _radRegularKey;
    private RadioButton _radSpecialKey;

    private TextField _txtRegularKey;
    private PopupMenu _popSpecialKey;

    public VirtualKeyboardDialog() {
        //public constructor
    }

    public void onDecoded() {
        //on decoded
        _chkShift = (CheckBox)this.getDescendantWithID(chkShiftModifier);
        _chkControl = (CheckBox)this.getDescendantWithID(chkControlModifier);
        _chkAlt = (CheckBox)this.getDescendantWithID(chkAltModifier);
        _chkMeta = (CheckBox)this.getDescendantWithID(chkMetaModifier);

        _radRegularKey = (RadioButton)this.getDescendantWithID(radRegularKey);
        _radSpecialKey = (RadioButton)this.getDescendantWithID(radSpecialKey);

        _txtRegularKey = (TextField)this.getDescendantWithID(txtRegularKey);
        _popSpecialKey = (PopupMenu)this.getDescendantWithID(popSpecialKeyMenu);

        for (int x=0; x < rfbProto.keyCodes.length; x++) {
            String name = rfbProto.keyNames[x];
            _popSpecialKey.addItem(name); //, evtVKBPopupMenu, rfbProto.keyCodes[x], name, this, 0);
        }
    }


    public boolean receiveEvent(Event event) {
        boolean handled = false;

        switch (event.type) {
            case evtKeyRadioChanged:
                if (_radRegularKey.getValue() == 0) {
                    _txtRegularKey.setVisible(false);
                    _popSpecialKey.setVisible(true);
                } else {
                    _txtRegularKey.setVisible(true);
                    _popSpecialKey.setVisible(false);
                }
                break;

            case kSendKeyEvent:
                int modifiers = 0;
                int keyCode = 0;
                if (_chkShift.getValue() != 0) modifiers += Event.EVENT_MODIFIER_CAPS;
                if (_chkControl.getValue() != 0) modifiers += Event.EVENT_MODIFIER_MENU_BUTTON;
                if (_chkAlt.getValue() != 0) modifiers += Event.EVENT_MODIFIER_ALT;
                if (_chkMeta.getValue() != 0) modifiers += Event.EVENT_MODIFIER_SYM;

                if (_radRegularKey.getValue() != 0) {
                    String keyStr = _txtRegularKey.toString();
                    if (keyStr.length() == 1)
                       keyCode = keyStr.charAt(0);
                } else {
                    keyCode = rfbProto.keyCodes[_popSpecialKey.getValue()];
                }

                if (keyCode != 0) {
                    try {
                        DEBUG.p("Sending Key: " + keyCode + " with modifiers: " + modifiers);
                        _rfb.sendKeyPressEvent(keyCode, modifiers);
                    } catch (IOException e) {}
                }
                break;
        }
        return handled || super.receiveEvent(event);
    }

    public static VirtualKeyboardDialog newVirtualKeyboard(rfbProto rfb) {
        VirtualKeyboardDialog d = (VirtualKeyboardDialog)Application.getCurrentApp().getResources().getDialog(kVirtualKeyboard);
        d._rfb = rfb;
        return d;
    }
}
