// Copyright (C) 2003 Jake Bordens.  All Rights Reserved.
// Copyright (C) 2002 Ultr@VNC Team Members. All Rights Reserved.
// Copyright (C) 2001,2002 HorizonLive.com, Inc.  All Rights Reserved.
// Copyright (C) 2002 Constantin Kaplinsky.  All Rights Reserved.
// Copyright (C) 1999 AT&T Laboratories Cambridge.  All Rights Reserved.

//  This is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 2 of the License, or
//  (at your option) any later version.
//
//  This software is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this software; if not, write to the Free Software
//  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307,
//  USA.

//
//  This file is part of
//         Hiptop VNC Client - a VNC client for the danger hiptop.
//

//-beta 2 released-
//version .12 - fixed EOF problem (-1 from stream) in readServerMessageType

package org.twodot.vnc;

import java.io.*;
import java.net.Socket;

import danger.app.*;
import danger.util.DEBUG;

//import danger.net.*;

class rfbProto {

    static final String versionMsg = "RFB 003.003\n";
    static final int ConnFailed = 0, NoAuth = 1, VncAuth = 2;
    static final int VncAuthOK = 0, VncAuthFailed = 1, VncAuthTooMany = 2;

    static final int FramebufferUpdate = 0, SetColourMapEntries = 1, Bell = 2,
    ServerCutText = 3, ResizeFrameBuffer = 4, PalmVNCReSizeFrameBuffer = 0xF;

    static final int SetPixelFormat = 0, FixColourMapEntries = 1,
    SetEncodings = 2, FramebufferUpdateRequest = 3, KeyEvent = 4,
    PointerEvent = 5, ClientCutText = 6, SetScaleFactor = 0xF;

    static final int EncodingRaw = 0, EncodingCopyRect = 1, EncodingRRE = 2,
    EncodingCoRRE = 4, EncodingHextile = 5, EncodingZlib = 6, EncodingTight = 7,
    EncodingZlibHex = 8, EncodingZRLE = 16, EncodingCache = 0xFFFF0000,
    EncodingCacheEnable = 0xFFFF0001, EncodingXOR_Zlib = 0xFFFF0002,
    EncodingXORMonoColor_Zlib = 0xFFFF0003, EncodingXORMultiColor_Zlib = 0xFFFF0004,
    EncodingSolidColor = 0xFFFF0005, EncodingXOREnable = 0xFFFF0006,
    EncodingCacheZip = 0xFFFF0007, EncodingSolMonoZip = 0xFFFF0008,
    EncodingXCursor = 0xFFFFFF10, EncodingRichCursor = 0xFFFFFF11,
    EncodingLastRect = 0xFFFFFF20, EncodingNewFBSize = 0xFFFFFF21;


    static final int HextileRaw = (1 << 0);
    static final int HextileBackgroundSpecified = (1 << 1);
    static final int HextileForegroundSpecified = (1 << 2);
    static final int HextileAnySubrects = (1 << 3);
    static final int HextileSubrectsColoured = (1 << 4);

    final static int TightExplicitFilter = 0x04;
    final static int TightFill = 0x08;
    final static int TightJpeg = 0x09;
    final static int TightMaxSubencoding = 0x09;
    final static int TightFilterCopy = 0x00;
    final static int TightFilterPalette = 0x01;
    final static int TightFilterGradient = 0x02;

    final static int EncodingCompressLevel0 = 0xFFFFFF00;
    final static int EncodingQualityLevel0 = 0xFFFFFFE0;

    final static int TightMinToCompress = 12;

    //KeyCode constants
    public static final int KEY_BACKSPACE = 0xff08;
    public static final int KEY_TAB = 0xff09;
    public static final int KEY_RETURN = 0xff0d;
    public static final int KEY_ENTER = 0xff0d;
    public static final int KEY_ESCAPE = 0xff1b;
    public static final int KEY_INSERT = 0xff63;
    public static final int KEY_DELETE = 0xffff;
    public static final int KEY_HOME = 0xff50;
    public static final int KEY_END = 0xff57;
    public static final int KEY_PAGE_UP = 0xff55;
    public static final int KEY_PAGE_DOWN = 0xff56;
    public static final int KEY_LEFT = 0xff51;
    public static final int KEY_UP = 0xff52;
    public static final int KEY_RIGHT = 0xff53;
    public static final int KEY_DOWN = 0xff54;
    public static final int KEY_F1 = 0xffbe;
    public static final int KEY_F2 = 0xffbf;
    public static final int KEY_F3 = 0xffc0;
    public static final int KEY_F4 = 0xffc1;
    public static final int KEY_F5 = 0xffc2;
    public static final int KEY_F6 = 0xffc3;
    public static final int KEY_F7 = 0xffc4;
    public static final int KEY_F8 = 0xffc5;
    public static final int KEY_F9 = 0xffc6;
    public static final int KEY_F10 = 0xffc7;
    public static final int KEY_F11 = 0xffc8;
    public static final int KEY_F12 = 0xffc9;
    public static final int KEY_SHIFT = 0xffe1;
    public static final int KEY_CONTROL = 0xffe3;
    public static final int KEY_META = 0xffe7;
    public static final int KEY_ALT = 0xffe9;

    public static final int[] keyCodes = {KEY_BACKSPACE, KEY_TAB, KEY_RETURN, KEY_ENTER, KEY_ESCAPE,
                                          KEY_INSERT, KEY_DELETE, KEY_HOME, KEY_END, KEY_PAGE_UP,
                                          KEY_PAGE_DOWN, KEY_LEFT, KEY_UP, KEY_RIGHT, KEY_DOWN,
                                          KEY_F1, KEY_F2, KEY_F3, KEY_F4, KEY_F5, KEY_F6, KEY_F7,
                                          KEY_F8, KEY_F9, KEY_F10, KEY_F11, KEY_F12};
    public static final String[] keyNames = {"Backspace", "Tab", "Return", "Enter", "Escape", "Insert",
                                             "Delete", "Home", "End", "Page Up", "Page Down", "Left", "Up",
                                             "Right", "Down", "F1", "F2", "F3", "F4", "F5", "F6", "F7", "F8",
                                             "F9", "F10", "F11", "F12"};

    String host;
    int port;
    Socket sock;
    DataInputStream is;
    OutputStream os;
    boolean inNormalProtocol = false;
    BufferedInputStream bis;

    //
    // Constructor.  Just make TCP connection to RFB server.
    //
    rfbProto(String h, int p) throws IOException {
        //v = v1;
        host = h;
        port = p;
        sock = new Socket(host, port);

        //is = new DataInputStream(new RingBufferInputStream(sock.getInputStream()));
        bis =new BufferedInputStream(sock.getInputStream());
        is = new DataInputStream(bis);
        os = sock.getOutputStream();
    }


    void close() {
        try {
            sock.close();
        } catch (Exception e) {
            e.printStackTrace();
        }
    }

    //
    // Read server's protocol version message
    //
    int serverMajor, serverMinor;
    void readVersionMsg() throws IOException {

        synchronized (is) {
            byte[] b = new byte[12];

            is.readFully(b);

            if ((b[0] != 'R') || (b[1] != 'F') || (b[2] != 'B') || (b[3] != ' ')
                    || (b[4] < '0') || (b[4] > '9') || (b[5] < '0') || (b[5] > '9')
                    || (b[6] < '0') || (b[6] > '9') || (b[7] != '.')
                    || (b[8] < '0') || (b[8] > '9') || (b[9] < '0') || (b[9] > '9')
                    || (b[10] < '0') || (b[10] > '9') || (b[11] != '\n')) {
                throw new IOException("Host " + host + " port " + port +
                        " is not an RFB server");
            }

            serverMajor = (b[4] - '0') * 100 + (b[5] - '0') * 10 + (b[6] - '0');
            serverMinor = (b[8] - '0') * 100 + (b[9] - '0') * 10 + (b[10] - '0');
        }
    }


    //
    // Write our protocol version message
    //
    void writeVersionMsg() throws IOException {
        synchronized (os) {
            byte[] b = new byte[12];
            versionMsg.getBytes(0, 12, b, 0);
            os.write(b);
        }
    }


    //
    // Find out the authentication scheme.
    //
    int readAuthScheme() throws IOException {
        synchronized (is) {
            int authScheme = is.readInt();

            switch (authScheme) {

                case ConnFailed:
                    int reasonLen = is.readInt();
                    byte[] reason = new byte[reasonLen];
                    is.readFully(reason);

                    throw new IOException(new String(reason, 0));

                case NoAuth:
                case VncAuth:
                    return authScheme;

                default:
                    throw new IOException("Unknown authentication scheme from RFB " +
                            "server " + authScheme);

            }
        }
    }


    //
    // Write the client initialisation message
    //
    void writeClientInit() throws IOException {
        synchronized (os) {
            //if (v.options.shareDesktop) {
            os.write(1);
            //} else {
            //  os.write(0);
            //}
            //v.options.disableShareDesktop();
        }
    }


    //
    // Read the server initialisation message
    //
    String desktopName;
    int framebufferWidth, framebufferHeight;
    int bitsPerPixel, depth;
    boolean bigEndian, trueColour;
    int redMax, greenMax, blueMax, redShift, greenShift, blueShift;

    void readServerInit() throws IOException {
        synchronized (is) {
            framebufferWidth = is.readUnsignedShort();
            framebufferHeight = is.readUnsignedShort();
            bitsPerPixel = is.readUnsignedByte();
            depth = is.readUnsignedByte();
            bigEndian = (is.readUnsignedByte() != 0);
            trueColour = (is.readUnsignedByte() != 0);
            redMax = is.readUnsignedShort();
            greenMax = is.readUnsignedShort();
            blueMax = is.readUnsignedShort();
            redShift = is.readUnsignedByte();
            greenShift = is.readUnsignedByte();
            blueShift = is.readUnsignedByte();
            byte[] pad = new byte[3];
            is.read(pad);
            int nameLength = is.readInt();
            byte[] name = new byte[nameLength];
            is.readFully(name);
            desktopName = new String(name, 0);

            inNormalProtocol = true;
        }
    }


    //
    // Read the server message type
    //
    int readServerMessageType() throws IOException {
        synchronized (is) {
            int msg = is.read();

            //handles a loss in signal
            if (msg == -1)
                throw new EOFException();

            return msg;
        }
    }


    //
    // Read a FramebufferUpdate message
    //
    int updateNRects;
    void readFramebufferUpdate() throws IOException {
        synchronized (is) {
            is.readByte();//pad
            updateNRects = is.readUnsignedShort();
        }
    }


    // Read a FramebufferUpdate rectangle header
    int updateRectX, updateRectY, updateRectW, updateRectH, updateRectEncoding;
    void readFramebufferUpdateRectHdr() throws IOException {
        synchronized (is) {
            updateRectX = is.readUnsignedShort();
            updateRectY = is.readUnsignedShort();
            updateRectW = is.readUnsignedShort();
            updateRectH = is.readUnsignedShort();
            updateRectEncoding = is.readInt();

            //SolMonoZip needs this check disabled.
            //if ((updateRectX + updateRectW > framebufferWidth) ||
            //(updateRectY + updateRectH > framebufferHeight)) {
            //  throw new IOException("Framebuffer update rectangle too large: " +
            //		    updateRectW + "x" + updateRectH + " at (" +
            //		    updateRectX + "," + updateRectY + ")");
            //}
        }
    }

    void readPalmVNCReSizeFrameBuffer() throws IOException {
        synchronized (is) {
            is.readUnsignedByte();
            updateRectW = is.readUnsignedShort();
            updateRectH = is.readUnsignedShort();
            framebufferWidth = is.readUnsignedShort();
            framebufferHeight = is.readUnsignedShort();
            is.readUnsignedShort();
        }
    }

    //
    // Read CopyRect source X and Y.
    //
    int copyRectSrcX, copyRectSrcY;
    void readCopyRect() throws IOException {
        synchronized (is) {
            copyRectSrcX = is.readUnsignedShort();
            copyRectSrcY = is.readUnsignedShort();
        }
    }


    //
    // Read a ServerCutText message
    //
    String readServerCutText() throws IOException {
        synchronized (is) {
            byte[] pad = new byte[3];
            is.read(pad);
            int len = is.readInt();
            byte[] text = new byte[len];
            is.readFully(text);
            return new String(text, 0);
        }
    }


    //
    // Write a FramebufferUpdateRequest message
    //
    void writeFramebufferUpdateRequest(int x, int y, int w, int h, boolean incremental)
            throws IOException {
        synchronized (os) {
            byte[] b = new byte[10];

            b[0] = (byte) FramebufferUpdateRequest;
            b[1] = (byte) (incremental ? 1 : 0);
            b[2] = (byte) ((x >> 8) & 0xff);
            b[3] = (byte) (x & 0xff);
            b[4] = (byte) ((y >> 8) & 0xff);
            b[5] = (byte) (y & 0xff);
            b[6] = (byte) ((w >> 8) & 0xff);
            b[7] = (byte) (w & 0xff);
            b[8] = (byte) ((h >> 8) & 0xff);
            b[9] = (byte) (h & 0xff);

            os.write(b);
        }
    }


    //
    // Write a SetPixelFormat message
    //
    void writeSetPixelFormat(int bitsPerPixel, int depth, boolean bigEndian,
                             boolean trueColour,
                             int redMax, int greenMax, int blueMax,
                             int redShift, int greenShift, int blueShift)
            throws IOException {
        synchronized (os) {
            byte[] b = new byte[20];

            b[0] = (byte) SetPixelFormat;
            b[4] = (byte) bitsPerPixel;
            b[5] = (byte) depth;
            b[6] = (byte) (bigEndian ? 1 : 0);
            b[7] = (byte) (trueColour ? 1 : 0);
            b[8] = (byte) ((redMax >> 8) & 0xff);
            b[9] = (byte) (redMax & 0xff);
            b[10] = (byte) ((greenMax >> 8) & 0xff);
            b[11] = (byte) (greenMax & 0xff);
            b[12] = (byte) ((blueMax >> 8) & 0xff);
            b[13] = (byte) (blueMax & 0xff);
            b[14] = (byte) redShift;
            b[15] = (byte) greenShift;
            b[16] = (byte) blueShift;

            os.write(b);
        }
    }

    //
    // Write a SetPixelFormat message
    //
    void writeSetScaleFactor(byte factor) throws IOException {

        synchronized (os) {
            byte[] b = new byte[4];

            b[0] = (byte) SetScaleFactor;
            b[1] = (byte) factor;

            os.write(b);
        }

    }


    //
    // Write a FixColourMapEntries message.  The values in the red, green and
    // blue arrays are from 0 to 65535.
    //
    /*
    void writeFixColourMapEntries(int firstColour, int nColours,
                                  int[] red, int[] green, int[] blue)
            throws IOException {
        synchronized (os) {
            byte[] b = new byte[6 + nColours * 6];

            b[0] = (byte) FixColourMapEntries;
            b[2] = (byte) ((firstColour >> 8) & 0xff);
            b[3] = (byte) (firstColour & 0xff);
            b[4] = (byte) ((nColours >> 8) & 0xff);
            b[5] = (byte) (nColours & 0xff);

            for (int i = 0; i < nColours; i++) {
                b[6 + i * 6] = (byte) ((red[i] >> 8) & 0xff);
                b[6 + i * 6 + 1] = (byte) (red[i] & 0xff);
                b[6 + i * 6 + 2] = (byte) ((green[i] >> 8) & 0xff);
                b[6 + i * 6 + 3] = (byte) (green[i] & 0xff);
                b[6 + i * 6 + 4] = (byte) ((blue[i] >> 8) & 0xff);
                b[6 + i * 6 + 5] = (byte) (blue[i] & 0xff);
            }

            os.write(b);
        }
    }
    */

    //
    // Write a SetEncodings message
    //
    void writeSetEncodings(int[] encs, int len) throws IOException {
        synchronized (os) {
            byte[] b = new byte[4 + 4 * len];

            b[0] = (byte) SetEncodings;
            b[2] = (byte) ((len >> 8) & 0xff);
            b[3] = (byte) (len & 0xff);

            for (int i = 0; i < len; i++) {
                b[4 + 4 * i] = (byte) ((encs[i] >> 24) & 0xff);
                b[5 + 4 * i] = (byte) ((encs[i] >> 16) & 0xff);
                b[6 + 4 * i] = (byte) ((encs[i] >> 8) & 0xff);
                b[7 + 4 * i] = (byte) (encs[i] & 0xff);
            }

            os.write(b);
        }
    }


    //
    // Write a ClientCutText message
    //
    void writeClientCutText(String text) throws IOException {
        synchronized (os) {
            byte[] b = new byte[8 + text.length()];

            b[0] = (byte) ClientCutText;
            b[4] = (byte) ((text.length() >> 24) & 0xff);
            b[5] = (byte) ((text.length() >> 16) & 0xff);
            b[6] = (byte) ((text.length() >> 8) & 0xff);
            b[7] = (byte) (text.length() & 0xff);

            text.getBytes(0, text.length(), b, 8);

            os.write(b);
        }
    }


    //
    // A buffer for putting pointer and keyboard events before being sent.  This
    // is to ensure that multiple RFB events generated from a single Java Event
    // will all be sent in a single network packet.  The maximum possible
    // length is 4 modifier down events, a single key event followed by 4
    // modifier up events i.e. 9 key events or 72 bytes.
    //
    byte[] eventBuf = new byte[72];
    int eventBufLen;


    //
    // Write a pointer event message.  We may need to send modifier key events
    // around it to set the correct modifier state.  Also buttons 2 and 3 are
    // represented as having ALT and META modifiers respectively.
    //

    int pointerMask = 0;

    void writePointerClick(int x, int y) throws IOException {
        writePointerClick(x, y, 1);
    }

    void writePointerClick(int x, int y, int btn) throws IOException {
        synchronized (eventBuf) {
            synchronized (os) {
                eventBufLen = 0;
                if (x < 0) x = 0;
                if (y < 0) y = 0;

                //mouse down
                eventBuf[eventBufLen++] = (byte) PointerEvent;
                eventBuf[eventBufLen++] = (byte) btn;
                eventBuf[eventBufLen++] = (byte) ((x >> 8) & 0xff);
                eventBuf[eventBufLen++] = (byte) (x & 0xff);
                eventBuf[eventBufLen++] = (byte) ((y >> 8) & 0xff);
                eventBuf[eventBufLen++] = (byte) (y & 0xff);

                //mouse up
                eventBuf[eventBufLen++] = (byte) PointerEvent;
                eventBuf[eventBufLen++] = (byte) 0;
                eventBuf[eventBufLen++] = (byte) ((x >> 8) & 0xff);
                eventBuf[eventBufLen++] = (byte) (x & 0xff);
                eventBuf[eventBufLen++] = (byte) ((y >> 8) & 0xff);
                eventBuf[eventBufLen++] = (byte) (y & 0xff);

                os.write(eventBuf, 0, eventBufLen);
            }
        }
    }


    //
    // Write a key event message.  We may need to send modifier key events
    // around it to set the correct modifier state.  Also we need to translate
    // from the Java key values to the X keysym values used by the RFB protocol.
    //
    /*
    void writeKeyEvent(Event evt) throws IOException {
        synchronized (os) {
            synchronized (eventBuf) {

                //we only want key events in here
                if (!evt.isKeyEvent()) return;

                int key = evt.what;
                boolean down = false;

                if ((evt.type == Event.EVENT_KEY_DOWN))
                    down = true;

                //for now, only accept real keypresses
                int keycode = 0;

                if (key >= 32 && key <= 256)
                    keycode = key;
                else if (key == 13)
                    keycode = 0xff0d;
                else if (key == 8)
                    keycode = 0xff08;

                eventBufLen = 0;

                writeModifierKeyEvents(evt.modifiers);
                writeKeyEvent(keycode, down);

                //
                // Always release all modifiers after an "up" event
                //

                if (!down) {
                    writeModifierKeyEvents(0);
                }

                os.write(eventBuf, 0, eventBufLen);
            }
        }
    }
    */

    //
    // Add a raw key event with the given X keysym to eventBuf.
    //

    void writeKeyEvent(int keysym, boolean down)
            throws IOException {
        synchronized (eventBuf) {
            eventBuf[eventBufLen++] = (byte) KeyEvent;
            eventBuf[eventBufLen++] = (byte) (down ? 1 : 0);
            eventBuf[eventBufLen++] = (byte) 0;
            eventBuf[eventBufLen++] = (byte) 0;
            eventBuf[eventBufLen++] = (byte) ((keysym >> 24) & 0xff);
            eventBuf[eventBufLen++] = (byte) ((keysym >> 16) & 0xff);
            eventBuf[eventBufLen++] = (byte) ((keysym >> 8) & 0xff);
            eventBuf[eventBufLen++] = (byte) (keysym & 0xff);
        }
    }


    void sendKeyPressEvent(int keysym, int modifiers) throws IOException {
        synchronized (eventBuf) {
            synchronized (os) {
            eventBufLen = 0;
                writeModifierKeyEvents(modifiers);
                writeKeyEvent(keysym, true);
                writeKeyEvent(keysym, false);
                writeModifierKeyEvents(0);
                os.write(eventBuf, 0, eventBufLen);
            }
        }
    }

    void sendModifiers(int modifiers) throws IOException {
        synchronized (eventBuf) {
            synchronized (os) {
                eventBufLen = 0;
                writeModifierKeyEvents(modifiers);
                os.write(eventBuf, 0, eventBufLen);
            }
        }
    }

    void sendKeyPressEvent(int keysym) throws IOException {
        synchronized (eventBuf) {
            synchronized (os) {
                eventBufLen = 0;
                writeKeyEvent(keysym, true);
                writeKeyEvent(keysym, false);
                os.write(eventBuf, 0, eventBufLen);
            }
        }
    }

    //
    // Write key events to set the correct modifier state.
    //

    int oldModifiers;

    void writeModifierKeyEvents(int newModifiers) throws IOException {
        //synchronized (this) {
        if ((newModifiers & Event.EVENT_MODIFIER_MENU_BUTTON) != (oldModifiers & Event.EVENT_MODIFIER_MENU_BUTTON))
            writeKeyEvent(0xffe3, (newModifiers & Event.EVENT_MODIFIER_MENU_BUTTON) != 0);

        if ((newModifiers & Event.EVENT_MODIFIER_CAPS) != (oldModifiers & Event.EVENT_MODIFIER_CAPS))
            writeKeyEvent(0xffe1, (newModifiers & Event.EVENT_MODIFIER_CAPS) != 0);

        //SYM==META
        if ((newModifiers & Event.EVENT_MODIFIER_SYM) != (oldModifiers & Event.EVENT_MODIFIER_SYM))
            writeKeyEvent(0xffe7, (newModifiers & Event.EVENT_MODIFIER_SYM) != 0);

        if ((newModifiers & Event.EVENT_MODIFIER_ALT) != (oldModifiers & Event.EVENT_MODIFIER_ALT))
            writeKeyEvent(0xffe9, (newModifiers & Event.EVENT_MODIFIER_ALT) != 0);

        //writeKeyEvent(0xffe1, (newModifiers & Event.EVENT_MODIFIER_ALT) != 0);

        oldModifiers = newModifiers;
        //}
    }

    int readCompactLen() throws IOException {
        synchronized (is) {
            byte len_byte;
            len_byte = (byte) (is.readUnsignedByte() & 0xFF);
            int compressedLen = ((int) len_byte) & 0x7F;
            if ((len_byte & 0x80) != 0) {
                len_byte = (byte) (is.readUnsignedByte() & 0xFF);
                compressedLen |= ((int) len_byte & 0x7F) << 7;
                if ((len_byte & 0x80) != 0) {
                    len_byte = (byte) (is.readUnsignedByte() & 0xFF);
                    compressedLen |= ((int) len_byte & 0xFF) << 14;
                }
            }
            //DEBUG.p("CompresedLEN:" + compressedLen);
            return compressedLen;
        }
    }

}
